from nuclia.sdk.search import AsyncNucliaSearch, NucliaSearch
from nucliadb_models.search import AskRequest, CustomPrompt
import pytest


def test_find(testing_config):
    search = NucliaSearch()
    results = search.find(query="Who is hedy Lamarr?")
    assert len(results.resources.keys()) == 2
    titles = [r.title for r in results.resources.values()]
    assert "Lamarr Lesson plan.pdf" in titles


def test_find_object(testing_config):
    search = NucliaSearch()
    results = search.find(query={"query": "Who is hedy Lamarr?"})
    assert len(results.resources.keys()) == 2
    titles = [r.title for r in results.resources.values()]
    assert "Lamarr Lesson plan.pdf" in titles


def test_ask(testing_config):
    search = NucliaSearch()
    results = search.ask(query="Who is Hedy Lamarr?")
    answer = results.answer.decode()
    assert "Lamarr" in answer


def test_ask_with_custom_prompt(testing_config):
    search = NucliaSearch()
    ask = AskRequest(
        query="Are Pepito Palotes and Hedy Lamarr friends?",
        prompt=CustomPrompt(
            system="Answer the question. If you don't know the anwser, write 'I don't know' and a list of bullet points with the reasons why you don't know the answer.",
            user="Based on this context {context}, answer the question {question}",
        ),
        generative_model="chatgpt-azure-4o",
    )
    results = search.ask(query=ask)
    answer = results.answer.decode()
    assert "I don't know" in answer, print(answer)


def test_ask_with_markdown_answer(testing_config):
    search = NucliaSearch()
    ask = AskRequest(
        query="Who is Hedy Lamarr and what did she do?",
        prompt=CustomPrompt(
            system="Answer the question and use lists as much as possible.",
            user="Based on this context {context}, answer the question {question}",
        ),
        generative_model="gemini-1-5-pro",
        prefer_markdown=True,
    )
    results = search.ask(query=ask)
    answer = results.answer.decode()
    markdown_keywords = ["**", "#", "1.", "*"]
    assert any([keyword in answer.lower() for keyword in markdown_keywords]), answer


def test_search(testing_config):
    search = NucliaSearch()
    results = search.search(query="Who is hedy Lamarr?")
    assert len(results.resources.keys()) == 2
    titles = [r.title for r in results.resources.values()]
    assert "Lamarr Lesson plan.pdf" in titles


def test_catalog(testing_config):
    search = NucliaSearch()
    results = search.catalog()
    assert len(results.resources.keys()) == 2


def test_search_object(testing_config):
    search = NucliaSearch()
    results = search.search(query={"query": "Who is hedy Lamarr?"})
    assert len(results.resources.keys()) == 2
    titles = [r.title for r in results.resources.values()]
    assert "Lamarr Lesson plan.pdf" in titles


def test_filters(testing_config):
    search = NucliaSearch()
    results = search.find(
        query="Who is hedy Lamarr?", filters=["/icon/application/pdf"]
    )
    assert len(results.resources.keys()) == 1


SCHEMA = {
    "name": "ClassificationReverse",
    "description": "Correctly extracted with all the required parameters with correct types",
    "parameters": {
        "$defs": {
            "Options": {
                "enum": ["SPORTS", "TECHNOLOGY"],
                "title": "Options",
                "type": "string",
            }
        },
        "properties": {
            "title": {"default": "label", "title": "Title", "type": "string"},
            "description": {
                "default": "Define labels to classify the subject of the document",
                "title": "Description",
                "type": "string",
            },
            "document_type": {
                "description": "Type of document, SPORT example: elections, Illa, POLITICAL example: football, TECHNOLOGY example: computer",
                "items": {"$ref": "#/$defs/Options"},
                "title": "Document Type",
                "type": "array",
            },
        },
        "required": ["document_type"],
        "type": "object",
    },
}


def test_ask_json(testing_config):
    search = NucliaSearch()
    results = search.ask_json(
        query="Who is hedy Lamarr?", filters=["/icon/application/pdf"], schema=SCHEMA
    )

    assert "TECHNOLOGY" in results.object["document_type"]


@pytest.mark.asyncio
async def test_ask_json_async(testing_config):
    search = AsyncNucliaSearch()
    results = await search.ask_json(
        query="Who is hedy Lamarr?", filters=["/icon/application/pdf"], schema=SCHEMA
    )

    assert "TECHNOLOGY" in results.object["document_type"]



from nuclia.sdk.agent import NucliaAgent
from nuclia.sdk.predict import NucliaPredict
from nuclia.sdk.process import NucliaProcessing


class NucliaNUA:
    def __init__(self):
        self.predict = NucliaPredict()
        self.process = NucliaProcessing()
        self.agent = NucliaAgent()



import json
import os
import warnings
from dataclasses import dataclass
from typing import Any, AsyncIterator, Dict, List, Optional, Union

from nucliadb_models.search import (
    AskRequest,
    AskResponseItem,
    CatalogRequest,
    Filter,
    FindRequest,
    KnowledgeboxFindResults,
    KnowledgeboxSearchResults,
    RagStrategies,
    RagImagesStrategies,
    Relations,
    SearchOptions,
    SearchRequest,
    SyncAskResponse,
    ChatModel,
)
from pydantic import ValidationError

from nuclia.data import get_async_auth, get_auth
from nuclia.decorators import kb, pretty
from nuclia.lib.kb import AsyncNucliaDBClient, NucliaDBClient
from nuclia.sdk.logger import logger
from nuclia.sdk.auth import AsyncNucliaAuth, NucliaAuth
from nuclia.sdk.resource import RagImagesStrategiesParse, RagStrategiesParse


@dataclass
class AskAnswer:
    answer: bytes
    object: Optional[Dict[str, Any]]
    learning_id: str
    relations_result: Optional[Relations]
    find_result: Optional[KnowledgeboxFindResults]
    citations: Optional[Dict[str, Any]]
    prequeries: Optional[Dict[str, KnowledgeboxFindResults]]
    timings: Optional[Dict[str, float]]
    tokens: Optional[Dict[str, int]]
    retrieval_best_matches: Optional[List[str]]
    status: Optional[str]
    prompt_context: Optional[list[str]]
    relations: Optional[Relations]
    predict_request: Optional[ChatModel]
    error_details: Optional[str]

    def __str__(self):
        if self.answer:
            return self.answer.decode()
        if self.object:
            return json.dumps(self.object)
        return ""


class NucliaSearch:
    """
    Perform search on a Knowledge Box.

    `find` and `search` accept the following parameters:
    - `json`: return results in JSON format
    - `yaml`: return results in YAML format
    """

    @property
    def _auth(self) -> NucliaAuth:
        auth = get_auth()
        return auth

    @kb
    @pretty
    def search(
        self,
        *,
        query: Union[str, SearchRequest] = "",
        filters: Optional[Union[List[str], List[Filter]]] = None,
        **kwargs,
    ) -> KnowledgeboxSearchResults:
        """
        Perform a search query.

        See https://docs.nuclia.dev/docs/api#tag/Search/operation/Search_Knowledge_Box_kb__kbid__search_post
        """
        ndb: NucliaDBClient = kwargs["ndb"]
        if isinstance(query, str):
            req = SearchRequest(query=query, filters=(filters or []), **kwargs)  # type: ignore
        elif isinstance(query, SearchRequest):
            req = query
        elif isinstance(query, dict):
            try:
                req = SearchRequest.model_validate(query)
            except ValidationError:
                logger.exception("Error validating query")
                raise
        else:
            raise Exception("Invalid Query either str or SearchRequest")

        return ndb.ndb.search(req, kbid=ndb.kbid)

    @kb
    @pretty
    def find(
        self,
        *,
        query: Union[str, FindRequest] = "",
        highlight: Optional[bool] = False,
        relations: Optional[bool] = False,
        filters: Optional[Union[List[str], List[Filter]]] = None,
        **kwargs,
    ):
        """
        Perform a find query.

        See https://docs.nuclia.dev/docs/api#tag/Search/operation/Find_Knowledge_Box_kb__kbid__find_post
        """
        ndb: NucliaDBClient = kwargs["ndb"]
        if isinstance(query, str) and highlight is not None:
            req = FindRequest(
                query=query,
                highlight=highlight,
                filters=filters or [],  # type: ignore
                **kwargs,
            )
        elif isinstance(query, FindRequest):
            req = query
        elif isinstance(query, dict):
            try:
                req = FindRequest.model_validate(query)
            except ValidationError:
                logger.exception("Error validating query")
                raise
        else:
            raise Exception("Invalid Query either str or FindRequest")

        if relations:
            req.features.append(SearchOptions.RELATIONS)

        return ndb.ndb.find(req, kbid=ndb.kbid)

    @kb
    @pretty
    def catalog(
        self,
        *,
        query: Union[str, CatalogRequest] = "",
        filters: Optional[Union[List[str], List[Filter]]] = None,
        **kwargs,
    ):
        """
        Perform a catalog query.

        See https://docs.nuclia.dev/docs/api#tag/Search/operation/catalog_post_kb__kbid__catalog_post
        """
        ndb: NucliaDBClient = kwargs["ndb"]
        if isinstance(query, str):
            req = CatalogRequest(
                query=query,
                filters=filters or [],  # type: ignore
                **kwargs,
            )
        elif isinstance(query, CatalogRequest):
            req = query
        elif isinstance(query, dict):
            try:
                req = CatalogRequest.model_validate(query)
            except ValidationError:
                logger.exception("Error validating query")
                raise
        else:
            raise Exception("Invalid Query either str or CatalogRequest")

        return ndb.ndb.catalog(req, kbid=ndb.kbid)

    @kb
    def ask(
        self,
        *,
        query: Union[str, dict, AskRequest],
        filters: Optional[Union[List[str], List[Filter]]] = None,
        rag_strategies: Optional[list[RagStrategies]] = None,
        rag_images_strategies: Optional[list[RagImagesStrategies]] = None,
        **kwargs,
    ):
        """
        Answer a question.

        See https://docs.nuclia.dev/docs/api#tag/Search/operation/Ask_Knowledge_Box_kb__kbid__ask_post
        """
        ndb: NucliaDBClient = kwargs["ndb"]
        if isinstance(query, str):
            req = AskRequest(
                query=query,
                filters=filters or [],  # type: ignore
                **kwargs,
            )
            if rag_strategies is not None:
                req.rag_strategies = RagStrategiesParse.model_validate(
                    {"rag_strategies": rag_strategies}
                ).rag_strategies
            if rag_images_strategies is not None:
                req.rag_images_strategies = RagImagesStrategiesParse.model_validate(
                    {"rag_images_strategies": rag_images_strategies}
                ).rag_images_strategies
        elif isinstance(query, dict):
            try:
                req = AskRequest.model_validate(query)
            except ValidationError:
                logger.exception("Error validating query")
                raise
        elif isinstance(query, AskRequest):
            req = query
        else:
            raise ValueError("Invalid query type. Must be str, dict or AskRequest.")

        ask_response: SyncAskResponse = ndb.ndb.ask(kbid=ndb.kbid, content=req)

        result = AskAnswer(
            answer=ask_response.answer.encode(),
            learning_id=ask_response.learning_id,
            relations_result=ask_response.relations,
            find_result=ask_response.retrieval_results,
            prequeries=ask_response.prequeries,
            retrieval_best_matches=[
                best.id for best in ask_response.retrieval_best_matches
            ],
            citations=ask_response.citations,
            timings=None,
            tokens=None,
            object=ask_response.answer_json,
            status=ask_response.status,
            error_details=ask_response.error_details,
            predict_request=ChatModel.model_validate(ask_response.predict_request)
            if ask_response.predict_request is not None
            else None,
            relations=ask_response.relations,
            prompt_context=ask_response.prompt_context,
        )

        if ask_response.prompt_context:
            result.prompt_context = ask_response.prompt_context
        if ask_response.metadata is not None:
            if ask_response.metadata.timings is not None:
                result.timings = ask_response.metadata.timings.model_dump()
            if ask_response.metadata.tokens is not None:
                result.tokens = ask_response.metadata.tokens.model_dump()

        return result

    @kb
    def ask_json(
        self,
        schema: Union[str, Dict[str, Any]],
        query: Union[str, dict, AskRequest, None] = None,
        filters: Optional[Union[List[str], List[Filter]]] = None,
        **kwargs,
    ):
        """
        Answer a question.

        See https://docs.nuclia.dev/docs/api#tag/Search/operation/Ask_Knowledge_Box_kb__kbid__ask_post
        """
        ndb: NucliaDBClient = kwargs["ndb"]
        if isinstance(schema, str):
            if os.path.exists(schema):
                with open(schema, "r") as json_file_handler:
                    try:
                        schema_json = json.load(json_file_handler)
                    except Exception:
                        logger.exception("File format is not JSON")
                        return
            else:
                logger.exception("File not found")
                return
        else:
            schema_json = schema

        if isinstance(query, str):
            req = AskRequest(
                query=query,
                answer_json_schema=schema_json,
            )
            if filters is not None:
                req.filters = filters
        elif isinstance(query, dict):
            try:
                req = AskRequest.model_validate(query)
            except ValidationError:
                logger.exception("Error validating query")
                raise
        elif isinstance(query, AskRequest):
            req = query
        elif query is None:
            req = AskRequest(
                query="",
                answer_json_schema=schema_json,
            )
            if filters is not None:
                req.filters = filters
        else:
            raise ValueError("Invalid query type. Must be str, dict or AskRequest.")
        ask_response: SyncAskResponse = ndb.ndb.ask(kbid=ndb.kbid, content=req)

        result = AskAnswer(
            answer=ask_response.answer.encode(),
            learning_id=ask_response.learning_id,
            relations_result=ask_response.relations,
            find_result=ask_response.retrieval_results,
            prequeries=ask_response.prequeries,
            retrieval_best_matches=[
                best.id for best in ask_response.retrieval_best_matches
            ],
            citations=ask_response.citations,
            timings=None,
            tokens=None,
            object=ask_response.answer_json,
            status=ask_response.status,
            error_details=ask_response.error_details,
            predict_request=ChatModel.model_validate(ask_response.predict_request)
            if ask_response.predict_request is not None
            else None,
            relations=ask_response.relations,
            prompt_context=ask_response.prompt_context,
        )
        if ask_response.metadata is not None:
            if ask_response.metadata.timings is not None:
                result.timings = ask_response.metadata.timings.model_dump()
            if ask_response.metadata.tokens is not None:
                result.tokens = ask_response.metadata.tokens.model_dump()
        return result


class AsyncNucliaSearch:
    """
    Perform search on a Knowledge Box.

    `find` and `search` accept the following parameters:
    - `json`: return results in JSON format
    - `yaml`: return results in YAML format
    """

    @property
    def _auth(self) -> AsyncNucliaAuth:
        auth = get_async_auth()
        return auth

    @kb
    @pretty
    async def search(
        self,
        *,
        query: Union[str, SearchRequest] = "",
        filters: Optional[List[str]] = None,
        **kwargs,
    ):
        """
        Perform a search query.

        See https://docs.nuclia.dev/docs/api#tag/Search/operation/Search_Knowledge_Box_kb__kbid__search_post
        """
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        if isinstance(query, str):
            req = SearchRequest(query=query, filters=(filters or []), **kwargs)  # type: ignore
        elif isinstance(query, SearchRequest):
            req = query
        elif isinstance(query, dict):
            try:
                req = SearchRequest.model_validate(query)
            except ValidationError:
                logger.exception("Error validating query")
                raise
        else:
            raise Exception("Invalid Query either str or SearchRequest")

        return await ndb.ndb.search(req, kbid=ndb.kbid)

    @kb
    @pretty
    async def find(
        self,
        *,
        query: Union[str, FindRequest] = "",
        highlight: Optional[bool] = False,
        relations: Optional[bool] = False,
        filters: Optional[List[str]] = None,
        **kwargs,
    ) -> KnowledgeboxFindResults:
        """
        Perform a find query.

        See https://docs.nuclia.dev/docs/api#tag/Search/operation/Find_Knowledge_Box_kb__kbid__find_post
        """

        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        if isinstance(query, str) and highlight is not None:
            req = FindRequest(
                query=query, highlight=highlight, filters=(filters or []), **kwargs
            )
        elif isinstance(query, FindRequest):
            req = query
        elif isinstance(query, dict):
            try:
                req = FindRequest.model_validate(query)
            except ValidationError:
                logger.exception("Error validating query")
                raise
        else:
            raise Exception("Invalid Query either str or FindRequest")

        if relations:
            req.features.append(SearchOptions.RELATIONS)

        return await ndb.ndb.find(req, kbid=ndb.kbid)

    @kb
    @pretty
    async def catalog(
        self,
        *,
        query: Union[str, CatalogRequest] = "",
        filters: Optional[Union[List[str], List[Filter]]] = None,
        **kwargs,
    ):
        """
        Perform a catalog query.

        See https://docs.nuclia.dev/docs/api#tag/Search/operation/catalog_post_kb__kbid__catalog_post
        """
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        if isinstance(query, str):
            req = CatalogRequest(
                query=query,
                filters=filters or [],  # type: ignore
                **kwargs,
            )
        elif isinstance(query, CatalogRequest):
            req = query
        elif isinstance(query, dict):
            try:
                req = CatalogRequest.model_validate(query)
            except ValidationError:
                logger.exception("Error validating query")
                raise
        else:
            raise Exception("Invalid Query either str or CatalogRequest")

        return await ndb.ndb.catalog(req, kbid=ndb.kbid)

    @kb
    async def ask(
        self,
        *,
        query: Union[str, dict, AskRequest],
        filters: Optional[List[str]] = None,
        timeout: int = 100,
        **kwargs,
    ):
        """
        Answer a question.

        See https://docs.nuclia.dev/docs/api#tag/Search/operation/Ask_Knowledge_Box_kb__kbid__ask_post
        """
        ndb: AsyncNucliaDBClient = kwargs["ndb"]

        if isinstance(query, str):
            req = AskRequest(
                query=query,
                filters=filters or [],  # type: ignore
                **kwargs,
            )
        elif isinstance(query, dict):
            try:
                req = AskRequest.model_validate(query)
            except ValidationError:
                logger.exception("Error validating query")
                raise
        elif isinstance(query, AskRequest):
            req = query
        else:
            raise ValueError("Invalid query type. Must be str, dict or AskRequest.")
        ask_stream_response = await ndb.ask(req, timeout=timeout)
        result = AskAnswer(
            answer=b"",
            learning_id=ask_stream_response.headers.get("NUCLIA-LEARNING-ID", ""),
            relations_result=None,
            find_result=None,
            prequeries=None,
            retrieval_best_matches=None,
            citations=None,
            timings=None,
            tokens=None,
            object=None,
            status=None,
            error_details=None,
            predict_request=None,
            relations=None,
            prompt_context=None,
        )
        async for line in ask_stream_response.aiter_lines():
            try:
                ask_response_item = AskResponseItem.model_validate_json(line).item
            except Exception as e:
                warnings.warn(f"Failed to parse AskResponseItem: {e}. item: {line}")
                continue
            if ask_response_item.type == "answer":
                result.answer += ask_response_item.text.encode()
            elif ask_response_item.type == "answer_json":
                result.object = ask_response_item.object
            elif ask_response_item.type == "retrieval":
                result.find_result = ask_response_item.results
            elif ask_response_item.type == "relations":
                result.relations_result = ask_response_item.relations
            elif ask_response_item.type == "citations":
                result.citations = ask_response_item.citations
            elif ask_response_item.type == "metadata":
                if ask_response_item.timings:
                    result.timings = ask_response_item.timings.model_dump()
                if ask_response_item.tokens:
                    result.tokens = ask_response_item.tokens.model_dump()
            elif ask_response_item.type == "status":
                result.status = ask_response_item.status
            elif ask_response_item.type == "prequeries":
                result.prequeries = ask_response_item.results
            elif ask_response_item.type == "error":
                result.error_details = ask_response_item.error
            elif ask_response_item.type == "debug":
                result.prompt_context = ask_response_item.metadata.get("prompt_context")
                result.predict_request = ask_response_item.metadata.get(
                    "predict_request"
                )
            else:  # pragma: no cover
                warnings.warn(f"Unknown ask stream item type: {ask_response_item.type}")
        return result

    @kb
    async def ask_stream(
        self,
        *,
        query: Union[str, dict, AskRequest],
        filters: Optional[List[str]] = None,
        timeout: int = 100,
        **kwargs,
    ) -> AsyncIterator[AskResponseItem]:
        """
        Answer a question.

        See https://docs.nuclia.dev/docs/api#tag/Search/operation/Ask_Knowledge_Box_kb__kbid__ask_post
        """
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        if isinstance(query, str):
            req = AskRequest(
                query=query,
                filters=filters or [],  # type: ignore
            )
        elif isinstance(query, dict):
            try:
                req = AskRequest.model_validate(query)
            except ValidationError:
                logger.exception("Error validating query")
                raise
        elif isinstance(query, AskRequest):
            req = query
        else:
            raise ValueError("Invalid query type. Must be str, dict or AskRequest.")
        ask_stream_response = await ndb.ask(req, timeout=timeout)
        async for line in ask_stream_response.aiter_lines():
            try:
                ask_response_item = AskResponseItem.model_validate_json(line)
            except Exception as e:
                warnings.warn(f"Failed to parse AskResponseItem: {e}. item: {line}")
                continue
            yield ask_response_item

    @kb
    async def ask_json(
        self,
        *,
        query: Union[str, dict, AskRequest],
        schema: Dict[str, Any],
        filters: Optional[List[str]] = None,
        timeout: int = 100,
        **kwargs,
    ):
        """
        Answer a question.

        See https://docs.nuclia.dev/docs/api#tag/Search/operation/Ask_Knowledge_Box_kb__kbid__ask_post
        """
        ndb: AsyncNucliaDBClient = kwargs["ndb"]

        if isinstance(query, str):
            req = AskRequest(
                query=query,
                answer_json_schema=schema,
                filters=filters or [],  # type: ignore
            )
        elif isinstance(query, dict):
            try:
                req = AskRequest.model_validate(query)
            except ValidationError:
                logger.exception("Error validating query")
                raise
        elif isinstance(query, AskRequest):
            req = query
        else:
            raise ValueError("Invalid query type. Must be str, dict or AskRequest.")
        ask_stream_response = await ndb.ask(req, timeout=timeout)
        result = AskAnswer(
            answer=b"",
            learning_id=ask_stream_response.headers.get("NUCLIA-LEARNING-ID", ""),
            relations_result=None,
            find_result=None,
            prequeries=None,
            retrieval_best_matches=None,
            citations=None,
            timings=None,
            tokens=None,
            object=None,
            status=None,
            error_details=None,
            predict_request=None,
            relations=None,
            prompt_context=None,
        )
        async for line in ask_stream_response.aiter_lines():
            try:
                ask_response_item = AskResponseItem.model_validate_json(line).item
            except Exception as e:
                warnings.warn(f"Failed to parse AskResponseItem: {e}. item: {line}")
                continue
            if ask_response_item.type == "answer":
                result.answer += ask_response_item.text.encode()
            elif ask_response_item.type == "retrieval":
                result.find_result = ask_response_item.results
            elif ask_response_item.type == "answer_json":
                result.object = ask_response_item.object
            elif ask_response_item.type == "relations":
                result.relations_result = ask_response_item.relations
            elif ask_response_item.type == "citations":
                result.citations = ask_response_item.citations
            elif ask_response_item.type == "metadata":
                if ask_response_item.timings:
                    result.timings = ask_response_item.timings.model_dump()
                if ask_response_item.tokens:
                    result.tokens = ask_response_item.tokens.model_dump()
            elif ask_response_item.type == "status":
                result.status = ask_response_item.status
            elif ask_response_item.type == "prequeries":
                result.prequeries = ask_response_item.results
            elif ask_response_item.type == "error":
                result.error_details = ask_response_item.error
            elif ask_response_item.type == "debug":
                result.prompt_context = ask_response_item.metadata.get("prompt_context")
                result.predict_request = ask_response_item.metadata.get(
                    "predict_request"
                )
            else:  # pragma: no cover
                warnings.warn(f"Unknown ask stream item type: {ask_response_item.type}")
        return result



from nuclia.sdk.auth import NucliaAuth
from nuclia.tests.fixtures import TESTING_KB


def test_auth_user(testing_user: str):
    na = NucliaAuth()
    assert na._validate_user_token(testing_user)


def test_auth_kb(testing_kb: str):
    na = NucliaAuth()
    kbobj = na.validate_kb(TESTING_KB, testing_kb)
    assert kbobj
    assert kbobj.uuid
    assert kbobj.config
    assert kbobj.config.title


def test_auth_nua(testing_nua: str):
    na = NucliaAuth()
    client, account_type, account, region = na.validate_nua(testing_nua)
    assert client
    assert account_type
    assert account


def test_auth_pat(testing_config):
    na = NucliaAuth()
    token = na.create_personal_token(description="sdk test token", days=1, login=False)
    assert token
    tokens = na.list_personal_tokens()
    assert len([t.id for t in tokens if t.id == token.id]) == 1
    na.delete_personal_token(token_id=token.id)
    tokens = na.list_personal_tokens()
    assert len([t.id for t in tokens if t.id == token.id]) == 0



from nuclia.decorators import kb
from nuclia.lib.kb import NucliaDBClient, AsyncNucliaDBClient
from nuclia_models.events.remi import (
    AggregatedRemiScoreMetric,
    RemiQuery,
    RemiQueryResults,
    RemiQueryResultWithContext,
)
from nuclia_models.common.utils import Aggregation
from typing import Union, Optional
from datetime import datetime
from pydantic import TypeAdapter

WAIT_FOR_DOWNLOAD_TIMEOUT = 120


class NucliaRemi:
    @kb
    def query(
        self,
        *args,
        query: Union[
            dict,
            RemiQuery,
        ],
        **kwargs,
    ) -> RemiQueryResults:
        """
        Get a list of rag request that matches a remi scores query.

        :param query: RemiQuery
        """
        _query: RemiQuery
        if isinstance(query, dict):
            _query = RemiQuery.model_validate(query)
        else:
            _query = query

        ndb: NucliaDBClient = kwargs["ndb"]
        response = ndb.remi_query(query=_query)
        return RemiQueryResults.model_validate(response.json())

    @kb
    def get_event(
        self,
        *args,
        event_id: int,
        **kwargs,
    ) -> RemiQueryResultWithContext:
        """
        Get a rag request with full context and REMI scores.
        Intended for obtaining complete context for an item originating from a /remi/query.

        :param event_id: int
        """
        ndb: NucliaDBClient = kwargs["ndb"]
        response = ndb.get_remi_event(event_id=event_id)
        return RemiQueryResultWithContext.model_validate(response.json())

    @kb
    def get_scores(
        self,
        *args,
        starting_at: Union[datetime, str],
        to: Optional[Union[datetime, str]],
        aggregation: Union[Aggregation, str],
        **kwargs,
    ) -> list[AggregatedRemiScoreMetric]:
        ndb: NucliaDBClient = kwargs["ndb"]
        if isinstance(starting_at, str):
            starting_at = datetime.fromisoformat(starting_at)
        if isinstance(to, str):
            to = datetime.fromisoformat(to)
        if isinstance(aggregation, str):
            aggregation = Aggregation[aggregation.upper()]
        response = ndb.get_remi_scores(
            _from=starting_at, to=to, aggregation=aggregation
        )
        ta = TypeAdapter(list[AggregatedRemiScoreMetric])
        return ta.validate_python(response.json())


class AsyncNucliaRemi:
    @kb
    async def query(
        self,
        *args,
        query: Union[
            dict,
            RemiQuery,
        ],
        **kwargs,
    ) -> RemiQueryResults:
        """
        Get a list of rag request that matches a remi scores query.

        :param query: RemiQuery
        """
        _query: RemiQuery
        if isinstance(query, dict):
            _query = RemiQuery.model_validate(query)
        else:
            _query = query

        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        response = await ndb.remi_query(query=_query)
        return RemiQueryResults.model_validate(response.json())

    @kb
    async def get_event(
        self,
        *args,
        event_id: int,
        **kwargs,
    ) -> RemiQueryResultWithContext:
        """
        Get a rag request with full context and REMI scores.
        Intended for obtaining complete context for an item originating from a /remi/query.

        :param event_id: int
        """
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        response = await ndb.get_remi_event(event_id=event_id)
        return RemiQueryResultWithContext.model_validate(response.json())

    @kb
    async def get_scores(
        self,
        *args,
        starting_at: Union[datetime, str],
        to: Optional[Union[datetime, str]],
        aggregation: Union[Aggregation, str],
        **kwargs,
    ) -> list[AggregatedRemiScoreMetric]:
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        if isinstance(starting_at, str):
            starting_at = datetime.fromisoformat(starting_at)
        if isinstance(to, str):
            to = datetime.fromisoformat(to)
        if isinstance(aggregation, str):
            aggregation = Aggregation[aggregation.upper()]
        response = await ndb.get_remi_scores(
            _from=starting_at, to=to, aggregation=aggregation
        )
        ta = TypeAdapter(list[AggregatedRemiScoreMetric])
        return ta.validate_python(response.json())



from nuclia.decorators import kb
from nuclia.lib.kb import LogType, NucliaDBClient, AsyncNucliaDBClient
from nuclia_models.events.activity_logs import (  # type: ignore
    ActivityLogsQuery,
    ActivityLogsChatQuery,
    ActivityLogsSearchQuery,
    DownloadActivityLogsQuery,
    DownloadActivityLogsChatQuery,
    DownloadActivityLogsSearchQuery,
    DownloadFormat,
    EventType,
)
from nuclia.lib.models import (
    ActivityLogsOutput,
    ActivityLogsQueryResponse,
    DownloadRequestOutput,
)
from typing import Union
from time import monotonic, sleep
from nuclia.sdk.logger import logger

import asyncio

WAIT_FOR_DOWNLOAD_TIMEOUT = 120


class NucliaLogs:
    @kb
    def get(
        self, *args, type: Union[LogType, str], month: str, **kwargs
    ) -> list[list[str]]:
        """
        Download activity logs.

        :param type: VISITED, MODIFIED, DELETED, NEW, SEARCH, SUGGEST, INDEXED, CHAT, STARTED, STOPPED, PROCESSED
        :param month: YYYY-MM
        """
        if isinstance(type, str):
            type = LogType[type.upper()]

        ndb: NucliaDBClient = kwargs["ndb"]
        resp = ndb.logs(type=type, month=month)
        return resp

    @kb
    def query(
        self,
        *args,
        type: Union[EventType, str],
        query: Union[
            dict,
            ActivityLogsQuery,
            ActivityLogsSearchQuery,
            ActivityLogsChatQuery,
        ],
        **kwargs,
    ) -> ActivityLogsOutput:
        """
        Query activity logs.

        :param type: VISITED, MODIFIED, DELETED, NEW, SEARCH, SUGGEST, INDEXED, CHAT, STARTED, STOPPED, PROCESSED
        :param query: ActivityLogsQuery
        """
        _type = EventType[type.upper()] if isinstance(type, str) else type
        _query: Union[
            ActivityLogsQuery,
            ActivityLogsSearchQuery,
            ActivityLogsChatQuery,
        ]
        if isinstance(query, dict):
            if _type is EventType.CHAT:
                _query = ActivityLogsChatQuery.model_validate(query)
            elif type is EventType.SEARCH:
                _query = ActivityLogsSearchQuery.model_validate(query)
            else:
                _query = ActivityLogsQuery.model_validate(query)
        else:
            _query = query

        ndb: NucliaDBClient = kwargs["ndb"]
        response = ndb.logs_query(type=_type, query=_query)
        output: list[ActivityLogsQueryResponse] = []  # type: ignore
        for line in response.iter_lines():
            output.append(ActivityLogsQueryResponse.model_validate_json(line))
        return ActivityLogsOutput(
            data=output, has_more=response.headers["has-more"].lower() == "true"
        )

    @kb
    def download(
        self,
        *args,
        type: Union[EventType, str],
        query: Union[
            dict,
            DownloadActivityLogsQuery,
            DownloadActivityLogsSearchQuery,
            DownloadActivityLogsChatQuery,
        ],
        download_format: Union[DownloadFormat, str],
        wait: bool = False,
        **kwargs,
    ) -> DownloadRequestOutput:  # type: ignore
        """
        Download activity logs.

        :param type: SEARCH, CHAT, VISITED, MODIFIED, DELETED, NEW, STARTED, STOPPED, PROCESSED
        :param type: NDJSON, CSV
        :param query: DownloadActivityLogsQuery
        """
        _type = EventType[type.upper()] if isinstance(type, str) else type
        _format = (
            DownloadFormat[download_format.upper()]
            if isinstance(download_format, str)
            else download_format
        )

        _query: Union[
            dict,
            DownloadActivityLogsQuery,
            DownloadActivityLogsSearchQuery,
            DownloadActivityLogsChatQuery,
        ]
        if isinstance(query, dict):
            if _type is EventType.CHAT:
                _query = DownloadActivityLogsChatQuery.model_validate(query)
            elif type is EventType.SEARCH:
                _query = DownloadActivityLogsSearchQuery.model_validate(query)
            else:
                _query = DownloadActivityLogsQuery.model_validate(query)
        else:
            _query = query

        ndb: NucliaDBClient = kwargs["ndb"]
        response = ndb.logs_download(type=_type, query=_query, download_format=_format)

        download_request = DownloadRequestOutput.model_validate(response.json())  # type: ignore

        if not wait:
            return download_request
        if wait:
            logger.info("Waiting for the download to be generated")
            t0 = monotonic()
            while monotonic() - t0 < 120:
                response = ndb.get_download_request(
                    request_id=download_request.request_id
                )
                download_request = DownloadRequestOutput.model_validate(response.json())  # type: ignore
                if download_request.download_url is not None:
                    break
                sleep(5)
            return download_request

    @kb
    def download_status(
        self,
        request_id: str,
        **kwargs,
    ) -> DownloadRequestOutput:  # type: ignore
        """
        Get the current status of a download request.

        :param request_id: The ID of the download request.
        """
        ndb: NucliaDBClient = kwargs["ndb"]
        response = ndb.get_download_request(request_id=request_id)
        return DownloadRequestOutput.model_validate(response.json())  # type: ignore


class AsyncNucliaLogs:
    @kb
    async def get(
        self, *args, type: Union[LogType, str], month: str, **kwargs
    ) -> list[list[str]]:
        """
        Download activity logs.

        :param type: VISITED, MODIFIED, DELETED, NEW, SEARCH, SUGGEST, INDEXED, CHAT, STARTED, STOPPED, PROCESSED
        :param month: YYYY-MM
        """
        if isinstance(type, str):
            type = LogType[type.upper()]

        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        resp = await ndb.logs(type=type, month=month)
        return resp

    @kb
    async def query(
        self,
        *args,
        type: Union[EventType, str],
        query: Union[
            dict,
            ActivityLogsQuery,
            ActivityLogsSearchQuery,
            ActivityLogsChatQuery,
        ],
        **kwargs,
    ) -> ActivityLogsOutput:
        """
        Query activity logs.

        :param type: VISITED, MODIFIED, DELETED, NEW, SEARCH, SUGGEST, INDEXED, CHAT, STARTED, STOPPED, PROCESSED
        :param query: ActivityLogsQuery
        """
        _type = EventType[type.upper()] if isinstance(type, str) else type
        _query: Union[
            ActivityLogsQuery,
            ActivityLogsSearchQuery,
            ActivityLogsChatQuery,
        ]
        if isinstance(query, dict):
            if _type is EventType.CHAT:
                _query = ActivityLogsChatQuery.model_validate(query)
            elif type is EventType.SEARCH:
                _query = ActivityLogsSearchQuery.model_validate(query)
            else:
                _query = ActivityLogsQuery.model_validate(query)
        else:
            _query = query

        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        response = await ndb.logs_query(type=_type, query=_query)
        output: list[ActivityLogsQueryResponse] = []  # type: ignore
        for line in response.iter_lines():
            output.append(ActivityLogsQueryResponse.model_validate_json(line))
        return ActivityLogsOutput(
            data=output, has_more=response.headers["has-more"].lower() == "true"
        )

    @kb
    async def download(
        self,
        *args,
        type: Union[EventType, str],
        query: Union[
            dict,
            DownloadActivityLogsQuery,
            DownloadActivityLogsSearchQuery,
            DownloadActivityLogsChatQuery,
        ],
        download_format: Union[DownloadFormat, str],
        wait: bool = False,
        **kwargs,
    ) -> DownloadRequestOutput:  # type: ignore
        """
        Download activity logs.

        :param type: SEARCH, CHAT, VISITED, MODIFIED, DELETED, NEW, STARTED, STOPPED, PROCESSED
        :param type: NDJSON, CSV
        :param query: DownloadActivityLogsQuery
        """
        _type = EventType[type.upper()] if isinstance(type, str) else type
        _format = (
            DownloadFormat[download_format.upper()]
            if isinstance(download_format, str)
            else download_format
        )

        _query: Union[
            dict,
            DownloadActivityLogsQuery,
            DownloadActivityLogsSearchQuery,
            DownloadActivityLogsChatQuery,
        ]
        if isinstance(query, dict):
            if _type is EventType.CHAT:
                _query = DownloadActivityLogsChatQuery.model_validate(query)
            elif type is EventType.SEARCH:
                _query = DownloadActivityLogsSearchQuery.model_validate(query)
            else:
                _query = DownloadActivityLogsQuery.model_validate(query)
        else:
            _query = query

        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        response = await ndb.logs_download(
            type=_type, query=_query, download_format=_format
        )

        download_request = DownloadRequestOutput.model_validate(response.json())  # type: ignore

        if not wait:
            return download_request
        if wait:
            logger.info("Waiting for the download to be generated")
            t0 = monotonic()
            while monotonic() - t0 < 120:
                response = await ndb.get_download_request(
                    request_id=download_request.request_id
                )
                download_request = DownloadRequestOutput.model_validate(response.json())  # type: ignore
                if download_request.download_url is not None:
                    break
                await asyncio.sleep(5)
            return download_request

    @kb
    async def download_status(
        self,
        request_id: str,
        **kwargs,
    ) -> DownloadRequestOutput:  # type: ignore
        """
        Get the current status of a download request.

        :param request_id: The ID of the download request.
        """
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        response = await ndb.get_download_request(request_id=request_id)
        return DownloadRequestOutput.model_validate(response.json())  # type: ignore



from nuclia.sdk.kb import NucliaKB, AsyncNucliaKB
from nuclia.tests.fixtures import IS_PROD
from nuclia_models.events.remi import RemiQuery, ContextRelevanceQuery
from datetime import datetime
from nuclia_models.common.utils import Aggregation
import pytest


def test_remi_query_and_get_event(testing_config):
    if not IS_PROD:
        assert True
        return
    nkb = NucliaKB()
    remi_query = nkb.remi.query(
        query=RemiQuery(
            month="2024-10",
            context_relevance=ContextRelevanceQuery(
                value=0, operation="eq", aggregation="max"
            ),
        )
    )
    assert len(remi_query.data) == 10
    remi_event = nkb.remi.get_event(event_id=remi_query.data[0].id)
    assert remi_event.id == remi_query.data[0].id


def test_remi_scores(testing_config):
    if not IS_PROD:
        assert True
        return
    nkb = NucliaKB()
    remi_scores_data = nkb.remi.get_scores(
        starting_at=datetime(year=2024, month=10, day=1),
        to=datetime(year=2024, month=11, day=1),
        aggregation=Aggregation.DAY,
    )
    assert len(remi_scores_data) == 32


@pytest.mark.asyncio
async def test_remi_query_and_get_event_async(testing_config):
    if not IS_PROD:
        assert True
        return
    nkb = AsyncNucliaKB()
    remi_query = await nkb.remi.query(
        query=RemiQuery(
            month="2024-10",
            context_relevance=ContextRelevanceQuery(
                value=0, operation="eq", aggregation="max"
            ),
        )
    )
    assert len(remi_query.data) == 10
    remi_event = await nkb.remi.get_event(event_id=remi_query.data[0].id)
    assert remi_event.id == remi_query.data[0].id


@pytest.mark.asyncio
async def test_remi_scores_async(testing_config):
    if not IS_PROD:
        assert True
        return
    nkb = AsyncNucliaKB()
    remi_scores_data = await nkb.remi.get_scores(
        starting_at=datetime(year=2024, month=10, day=1),
        to=datetime(year=2024, month=11, day=1),
        aggregation=Aggregation.DAY,
    )
    assert len(remi_scores_data) == 32
